package ru.dbotthepony.kstarbound.client

import ru.dbotthepony.kstarbound.api.IStruct2f
import ru.dbotthepony.kstarbound.client.render.ChunkRenderer
import ru.dbotthepony.kstarbound.client.render.renderLayeredList
import ru.dbotthepony.kstarbound.world.Chunk
import ru.dbotthepony.kstarbound.world.IWorldChunkTuple
import ru.dbotthepony.kstarbound.world.MutableWorldChunkTuple
import ru.dbotthepony.kstarbound.world.World

class ClientWorldChunkTuple(
	world: World<*>,
	chunk: Chunk,
	top: IWorldChunkTuple?,
	left: IWorldChunkTuple?,
	right: IWorldChunkTuple?,
	bottom: IWorldChunkTuple?,

	val renderer: ChunkRenderer
) : MutableWorldChunkTuple(
	world,
	chunk,
	top,
	left,
	right,
	bottom,
)

class ClientWorld(val client: StarboundClient, seed: Long = 0L) : World<ClientWorldChunkTuple>(seed) {
	override fun tupleFactory(
		chunk: Chunk,
		top: IWorldChunkTuple?,
		left: IWorldChunkTuple?,
		right: IWorldChunkTuple?,
		bottom: IWorldChunkTuple?
	): ClientWorldChunkTuple {
		return ClientWorldChunkTuple(
			world = this,
			chunk = chunk,
			top = top,
			left = left,
			right = right,
			bottom = bottom,

			renderer = ChunkRenderer(client.gl, chunk)
		)
	}

	/**
	 * Отрисовывает этот мир с точки зрения [pos] в Starbound Units
	 *
	 * Все координаты "местности" сохраняются, поэтому, если отрисовывать слишком далеко от 0, 0
	 * то геометрия может начать искажаться из-за погрешности плавающей запятой
	 *
	 * Обрезает всю заведомо невидимую геометрию на основе аргументов mins и maxs (в пикселях)
	 */
	fun render(
		pos: IStruct2f,
		scale: Float = 1f,

		mins: IStruct2f,
		maxs: IStruct2f,
	) {
		val determineRenderers = ArrayList<ChunkRenderer>()

		for (chunk in chunkMap.values) {
			determineRenderers.add(chunk.renderer)
		}

		val renderList = ArrayList<ChunkRenderer>()

		for (renderer in determineRenderers) {
			renderList.add(renderer)
			renderer.autoBakeStatic()
		}

		renderLayeredList(client.gl.matrixStack, renderList)
	}
}
